"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const cdk_1 = require("@aws-cdk/cdk");
const aws_apigateway_1 = require("@aws-cdk/aws-apigateway");
const CM = require("@aws-cdk/aws-certificatemanager");
const CF = require("@aws-cdk/aws-cloudfront");
const CB = require("@aws-cdk/aws-codebuild");
const aws_codepipeline_1 = require("@aws-cdk/aws-codepipeline");
const Events = require("@aws-cdk/aws-events");
const IAM = require("@aws-cdk/aws-iam");
const aws_lambda_1 = require("@aws-cdk/aws-lambda");
const aws_logs_1 = require("@aws-cdk/aws-logs");
const S3 = require("@aws-cdk/aws-s3");
const lib_1 = require("./lib");
const lodash_1 = require("lodash");
class SeagullStack extends cdk_1.Stack {
    constructor(app, projectName, stackProps) {
        super(app, projectName, stackProps);
    }
    addLogGroup(logGroupName) {
        const retentionDays = Infinity;
        const props = { logGroupName, retentionDays, retainLogGroup: false };
        return new aws_logs_1.LogGroup(this, `${this.id}-${logGroupName}`, props);
    }
    addS3(bucketName, role, keepBucket = true) {
        const removalPolicy = keepBucket ? cdk_1.RemovalPolicy.Orphan : cdk_1.RemovalPolicy.Destroy;
        const s3Props = { bucketName, removalPolicy };
        const bucket = new S3.Bucket(this, `${this.id}-${bucketName}`, s3Props);
        role && bucket.grantReadWrite(role);
        return bucket;
    }
    addLambda(name, folder, role, env) {
        const lambdaName = `${this.id}-${name}`;
        const conf = {
            code: aws_lambda_1.Code.asset(`${folder}/.seagull/deploy`),
            description: 'universal route',
            environment: env,
            functionName: `${lambdaName}-handler`,
            handler: 'dist/assets/backend/lambda.handler',
            memorySize: 1536,
            role,
            runtime: aws_lambda_1.Runtime.NodeJS810,
            timeout: 300,
        };
        return new aws_lambda_1.Function(this, lambdaName, conf);
    }
    addUniversalApiGateway(apiGWName, lambda, stageName) {
        const name = `${this.id}-${apiGWName}`;
        const defaultIntegration = new aws_apigateway_1.LambdaIntegration(lambda);
        const conf = { binaryMediaTypes: ['*/*'], deployOptions: { stageName } };
        const apiGateway = new aws_apigateway_1.RestApi(this, name, conf);
        const proxy = apiGateway.root.addResource('{any+}');
        apiGateway.root.addMethod('GET', defaultIntegration);
        apiGateway.root.addMethod('POST', defaultIntegration);
        apiGateway.root.addMethod('DELETE', defaultIntegration);
        proxy.addMethod('GET', defaultIntegration);
        proxy.addMethod('POST', defaultIntegration);
        proxy.addMethod('DELETE', defaultIntegration);
        return apiGateway;
    }
    addIAMRole(roleName, principleName, actions) {
        const name = `${this.id}-${roleName}`;
        const roleParams = { assumedBy: new IAM.ServicePrincipal(principleName) };
        const role = new IAM.Role(this, name, roleParams);
        const policyStatement = new IAM.PolicyStatement();
        role.addToPolicy(policyStatement.addAllResources().addActions(...actions));
        return role;
    }
    addCloudfront(cfdName, props) {
        const name = `${this.id}-${cfdName}`;
        const domainName = lib_1.getApiGatewayDomain(props.apiGateway.url);
        const originPath = lib_1.getApiGatewayPath(props.apiGateway.url);
        const behaviors = [this.getDefaultBehavior()];
        const customOriginSource = { domainName };
        const errorPageConfig = props.errorBucket && this.getErrorPageConfig(props.errorBucket);
        const originConfigurations = lodash_1.compact([
            { behaviors, customOriginSource, originPath },
            errorPageConfig,
        ]);
        const errorConfig = this.getErrorConfig(500, 500, '/error.html', 60);
        const conf = {
            aliasConfiguration: props.aliasConfig,
            comment: this.id,
            defaultRootObject: '',
            errorConfigurations: [errorConfig],
            loggingConfig: props.logBucket ? { bucket: props.logBucket } : {},
            originConfigs: originConfigurations,
        };
        return new CF.CloudFrontWebDistribution(this, name, conf);
    }
    addPipeline(pipelineName) {
        const name = `${this.id}-${pipelineName}`;
        return new aws_codepipeline_1.Pipeline(this, name, { pipelineName: name });
    }
    getErrorPageConfig(bucket) {
        const errorBehavior = {
            allowedMethods: CF.CloudFrontAllowedMethods.ALL,
            compress: true,
            forwardedValues: { headers: ['authorization'], queryString: true },
            isDefaultBehavior: false,
            pathPattern: '/error*',
        };
        const s3OriginConfig = {
            s3BucketSource: bucket,
        };
        const errorPageConfig = {
            behaviors: [errorBehavior],
            s3OriginSource: s3OriginConfig,
        };
        return errorPageConfig;
    }
    getDefaultBehavior() {
        return {
            allowedMethods: CF.CloudFrontAllowedMethods.ALL,
            compress: true,
            forwardedValues: { headers: ['authorization'], queryString: true },
            isDefaultBehavior: true,
        };
    }
    getErrorConfig(code, responseCode, responsePagePath, cachingTtl) {
        return {
            errorCachingMinTtl: cachingTtl,
            errorCode: code,
            responseCode,
            responsePagePath,
        };
    }
    addSourceStage(name, config) {
        const stageName = name;
        const sourceName = `${this.id}-github-${name}`;
        const { atIndex, branch, owner, pipeline, poll, repo, oauthToken } = config;
        const stage = pipeline.addStage(stageName, { placement: { atIndex } });
        const stageConfig = {
            branch,
            oauthToken,
            outputArtifactName: name,
            owner,
            pollForSourceChanges: poll,
            repo,
            stage,
        };
        return new aws_codepipeline_1.GitHubSourceAction(this, sourceName, stageConfig);
    }
    addGenericBuildActionStage(name, config) {
        const stageName = name;
        const projectName = `${this.id}-project-${name}`;
        const { atIndex, pipeline } = config;
        const projectConfig = this.createProjectConfig(config);
        const additionalOutputArtifactNames = this.getAdditionalOutputArtifactNames(name, config);
        const project = new CB.PipelineProject(this, projectName, projectConfig);
        const stage = pipeline.addStage(stageName, { placement: { atIndex } });
        const stageConfig = {
            additionalInputArtifacts: config.additionalInputArtifacts,
            additionalOutputArtifactNames,
            inputArtifact: config.inputArtifact,
            outputArtifactName: name,
            project,
            stage,
        };
        return new CB.PipelineBuildAction(this, name, stageConfig);
    }
    addDeployActionStage(name, config) {
        const stageName = name;
        const projectName = `${this.id}-project-${name}`;
        const { atIndex, pipeline } = config;
        const projectConfig = this.createProjectConfig(config);
        const additionalOutputArtifactNames = this.getAdditionalOutputArtifactNames(name, config);
        const project = new CB.PipelineProject(this, projectName, projectConfig);
        const stage = pipeline.addStage(stageName, { placement: { atIndex } });
        const stageConfig = {
            additionalInputArtifacts: config.additionalInputArtifacts,
            additionalOutputArtifactNames,
            inputArtifact: config.inputArtifact,
            outputArtifactName: name,
            project,
            stage,
        };
        return new CB.PipelineBuildAction(this, name, stageConfig);
    }
    addTestAction(name, stage, config) {
        const projectName = `${this.id}-project-${name}`;
        const projectConfig = this.createProjectConfig(config);
        const additionalOutputArtifactNames = this.getAdditionalOutputArtifactNames(name, config);
        const project = new CB.PipelineProject(this, projectName, projectConfig);
        const stageConfig = {
            additionalInputArtifacts: config.additionalInputArtifacts,
            additionalOutputArtifactNames,
            inputArtifact: config.inputArtifact,
            outputArtifactName: name,
            project,
            stage,
        };
        return new CB.PipelineBuildAction(this, name, stageConfig);
    }
    addBuildActionStage(name, config, testConfig, workerCount) {
        const stage = config.pipeline.addStage(name, {
            placement: { atIndex: config.atIndex },
        });
        const parallelBuildActions = workerCount;
        const actions = [...Array(parallelBuildActions).keys()].map(current => this.addBuildAction(name, stage, config, {
            count: parallelBuildActions,
            current,
        }));
        this.addTestAction('test', stage, testConfig);
        return actions;
    }
    addBuildAction(name, stage, config, workerMeta) {
        const projectName = `${this.id}-project-${name}-${workerMeta.current}`;
        config.env.variables.BUILD_WORKER_ID = workerMeta.current.toString();
        config.env.variables.BUILD_WORKER_COUNT = workerMeta.count.toString();
        const projectConfig = this.createProjectConfig(config);
        const project = new CB.PipelineProject(this, projectName, projectConfig);
        const stageConfig = {
            additionalInputArtifacts: config.additionalInputArtifacts,
            additionalOutputArtifactNames: [`dist${workerMeta.current}`],
            inputArtifact: config.inputArtifact,
            outputArtifactName: `build${workerMeta.current}`,
            project,
            stage,
        };
        const action = new CB.PipelineBuildAction(this, `${name}-${workerMeta.current}`, stageConfig);
        return action;
    }
    addSecretParam(name, ssmParameter) {
        const secretConfig = { ssmParameter };
        return new cdk_1.SecretParameter(this, name, secretConfig);
    }
    addNewCert(name, domains) {
        const domainName = domains[0];
        const altNames = domains.slice(1);
        const subjectAlternativeNames = altNames.length > 0 ? altNames : undefined;
        const props = { domainName, subjectAlternativeNames };
        return new CM.Certificate(this, `${this.id}-${name}`, props);
    }
    importS3(bucketName, role) {
        const name = `${this.id}-${bucketName}`;
        const bucketArn = `arn:aws:s3:::${bucketName}`;
        const bucket = S3.Bucket.import(this, name, { bucketArn });
        role && bucket.grantReadWrite(role);
        return bucket;
    }
    addEventRule(rule, target) {
        const name = rule.path.split('/').pop() || 'rootPath';
        const schedule = {
            scheduleExpression: rule.cron,
        };
        const eventRule = new Events.EventRule(this, name, schedule);
        eventRule.addTarget(target, { jsonTemplate: `{"path":"${rule.path}"}` });
        return eventRule;
    }
    getLogBucketConfig(name) {
        const bucket = name ? this.addS3(name) : false;
        return bucket ? { bucket } : {};
    }
    createProjectConfig(config) {
        const allowedComputeTypeSizes = ['SMALL', 'MEDIUM', 'LARGE'];
        const { build, env, install, postBuild, role } = config;
        const buildImage = CB.LinuxBuildImage.UBUNTU_14_04_NODEJS_8_11_0;
        const computeType = allowedComputeTypeSizes.includes(config.computeTypeSize)
            ? `BUILD_GENERAL1_${config.computeTypeSize}`
            : CB.ComputeType.Small;
        const phases = { build, install, post_build: postBuild };
        return {
            buildSpec: {
                artifacts: config.outputArtifacts,
                env,
                phases,
                version: '0.2',
            },
            environment: { buildImage, computeType },
            environmentVariables: lib_1.mapEnvironmentVariables(env.variables),
            role,
        };
    }
    getAdditionalOutputArtifactNames(name, config) {
        if (config.outputArtifacts) {
            return Object.keys(config.outputArtifacts['secondary-artifacts'])
                .filter(key => key !== name)
                .map(key => config.outputArtifacts['secondary-artifacts'][key].name);
        }
        return [];
    }
}
exports.SeagullStack = SeagullStack;
//# sourceMappingURL=seagull_stack.js.map