"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const aws_cdk_1 = require("aws-cdk");
const aws_sdk_handler_1 = require("../aws_sdk_handler");
const lib = require("../lib");
const seagull_app_1 = require("../seagull_app");
const set_aws_credentials_1 = require("../set_aws_credentials");
class SeagullPipeline {
    constructor(props) {
        this.appPath = props.appPath;
        this.branch = props.branch;
        this.computeTypeSize = props.computeTypeSize;
        this.stage = props.stage;
        this.poll = props.poll;
        this.profile = props.profile;
        this.owner = props.owner;
        this.region = props.region;
        this.repository = props.repository;
        this.ssmParam = props.ssmParameter;
        this.githubToken = props.githubToken;
        this.buildWorkers = props.buildWorkers;
        const propsSSMHandler = props.handlers && props.handlers.ssmHandler;
        this.ssm = propsSSMHandler || new aws_sdk_handler_1.SSMHandler();
        this.excludedPages = props.excludedPages;
        this.actions = [
            'cloudformation:*',
            'cloudfront:*',
            'iam:*',
            's3:*',
            'ssm:GetParameters',
            'apigateway:*',
            'lambda:*',
            'logs:*',
            'events:*',
        ];
    }
    createPipeline() {
        return __awaiter(this, void 0, void 0, function* () {
            set_aws_credentials_1.setCredsByProfile(this.profile);
            const isTest = this.stage === 'test';
            const suffix = `${isTest ? `-${this.branch}-test` : ''}-ci`;
            const pkgJson = require(`${this.appPath}/package.json`);
            const name = `${pkgJson.name}${suffix}`.replace(/[^0-9A-Za-z-]/g, '');
            const sdk = new aws_cdk_1.SDK({});
            const account = yield sdk.defaultAccount();
            const stackProps = { env: { account, region: this.region } };
            const props = { projectName: name, sdk, stackProps };
            const actions = this.actions;
            const pipelineApp = new seagull_app_1.SeagullApp(props);
            const stack = pipelineApp.stack;
            const principal = 'codebuild.amazonaws.com';
            const role = stack.addIAMRole('role', principal, actions);
            const pipeline = stack.addPipeline('pipeline');
            const token = this.githubToken;
            const tokenName = token ? `${name}-github` : this.ssmParam || undefined;
            const secretParams = { ssmHandler: this.ssm, token, tokenName };
            const ssmSecret = yield aws_sdk_handler_1.handleSSMSecret(secretParams);
            const gitDataProps = {
                branch: this.branch,
                owner: this.owner,
                pkg: pkgJson,
                repo: this.repository,
                secretParameter: ssmSecret.secret,
            };
            const gitData = lib.getGitData(gitDataProps);
            const pipelineDomain = `https://${this.region}.console.aws.amazon.com`;
            const pipelinePath = `/codesuite/codepipeline/pipelines/${pipeline.id}/view`;
            const pipelineLink = `${pipelineDomain}${pipelinePath}`;
            const stageConfigParams = {
                branch: gitData.branch,
                buildWorkers: this.buildWorkers,
                computeTypeSize: this.computeTypeSize,
                excludedPages: this.excludedPages,
                owner: gitData.owner,
                pipeline,
                pipelineLink,
                poll: this.poll,
                repo: gitData.repo,
                role,
                ssmSecret,
                stage: this.stage,
            };
            lib.addPipelineStages(stack, stageConfigParams);
            return pipelineApp;
        });
    }
    deployPipeline() {
        return __awaiter(this, void 0, void 0, function* () {
            const pipeline = yield this.createPipeline();
            yield pipeline.deployStack();
        });
    }
    destroyPipeline() {
        return __awaiter(this, void 0, void 0, function* () {
            const pipeline = yield this.createPipeline();
            yield pipeline.destroyStack();
        });
    }
}
exports.SeagullPipeline = SeagullPipeline;
//# sourceMappingURL=seagull_pipeline.js.map